# -*- coding: utf-8 -*-
"""
@author: hubo
@project: bb-py
@file: _redis.py
@time: 2023/11/10 16:06
@desc:
"""
import redis
from .util.singleton import Singleton


@Singleton
class _Redis:
    def __init__(self):
        """
        初始化redis
        :param url:类似这样的格式redis://user:password@localhost:6379/db
        """
        self.r = None

    def enabled_from(self, url):
        if self.r is None and url:
            self.r = redis.from_url(url)
        return self.r is not None

    def set(self, key, value, expire_in):
        self.r.set(key, value)
        self.r.expireat(key, expire_in)

    def set_expire_in(self, key, expire_in):
        self.r.expireat(key, expire_in)

    def get(self, key):
        return self.r.get(key)

    def delete(self, key):
        self.r.delete(key)




from time import sleep

import pytest
from fastapi import HTTPException

from fastoauth import OAuth2


def test_create_access_token():
    access_token = OAuth2.create_access_token({'user_id': 1, 'user_name': 'foo', 'tenant_id': 23})
    print(f'access_token: {access_token}')


def test_get_current_user():
    # 注意，测试用的SECRET_KEY，请勿用于生产环境
    OAuth2.SECRET_KEY = 'c1cc1a59d0d64ea04c4d62bc738dee398cd78348d18ef57b66ebad5b88e39d66'
    OAuth2.EXPIRE_SECONDS = 60 * 60 * 24 * 365
    # one_year_token = OAuth2.create_access_token({'user_id': 1, 'user_name': 'foo', 'tenant_id': 23})
    # print(f'一年有效期的token: {one_year_token}')
    one_year_token = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1c2VyX2lkIjoxLCJ1c2VyX25hbWUiOiJmb28iLCJ0ZW5hbnRfaWQiOjIzLCJleHAiOjE3NDM5NjE0NTh9.muCjjUvUBtTn418C8plTlZF2eWOQrOLLMLroSDSktR0'
    # 得到之前加密的字典数据
    print(OAuth2.get_token_user(one_year_token))


def test_refresh_token():
    # 注意，测试用的SECRET_KEY，请勿用于生产环境
    OAuth2.SECRET_KEY = 'c1cc1a59d0d64ea04c4d62bc738dee398cd78348d18ef57b66ebad5b88e39d66'
    OAuth2.EXPIRE_SECONDS = 60 * 60 * 24 * 2
    # access_token = OAuth2.create_access_token({'user_id': 1, 'user_name': 'foo', 'tenant_id': 23})
    # print(f'access_token: {access_token}')
    # 之前运行出来的一年有效期的refresh_token
    refresh_token = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1c2VyX2lkIjoxLCJ1c2VyX25hbWUiOiJmb28iLCJ0ZW5hbnRfaWQiOjIzLCJleHAiOjE3NDM5NjE0NTgsIl9yZWZyZXNoIjp0cnVlfQ.aDR-Pfw_m4clsBCt0KBQMZkml-790Jrg-t7eA0ZQkpQ'
    print(OAuth2.refresh_token(refresh_token))


def test_redis():
    # 需要先启动redis
    OAuth2.REDIS_URL = 'redis://localhost:6379/oauth2'
    indata = {'user_id': 1, 'user_name': 'aaaa', 'tenant_id': 23}
    access_token = OAuth2.create_access_token(indata)
    print(f'access_token: {access_token}')
    outdata = OAuth2.get_token_user(access_token.access_token)
    print(outdata)
    assert indata['user_id'] == outdata['user_id']
    # 模拟refresh
    refreshed_token = OAuth2.refresh_token(access_token.refresh_token)
    print(f'refreshed token: {refreshed_token}')
    # 模拟logout，测试access_token是否删除
    with pytest.raises(HTTPException) as ex:
        OAuth2.remove_token(access_token.access_token)
        OAuth2.get_token_user(access_token.access_token)  # 异常
    assert ex.value.status_code == 401
    assert ex.value.detail == 'Could not validate credentials'
    # 模拟logout，测试refresh_token是否删除
    with pytest.raises(HTTPException) as ex:
        OAuth2.remove_token(refreshed_token.access_token)
        OAuth2.refresh_token(refreshed_token.refresh_token)  # 异常
    assert ex.value.status_code == 401
    assert ex.value.detail == 'Could not validate credentials'
    print(f'token removed')


def test_redis_timeout():
    # 需要先启动redis
    OAuth2.REDIS_URL = 'redis://localhost:6379/oauth2'
    OAuth2.EXPIRE_SECONDS = 5
    indata = {'user_id': 1, 'user_name': 'aaaa', 'tenant_id': 23}
    access_token = OAuth2.create_access_token(indata)
    print(f'access_token: {access_token}')
    sleep(OAuth2.EXPIRE_SECONDS + 1)
    with pytest.raises(HTTPException) as ex:
        OAuth2.get_token_user(access_token.access_token)        # 异常
    assert ex.value.status_code == 401
    assert ex.value.detail == 'Could not validate credentials'
    print(f'token expired: {access_token}')




# -*- coding: utf-8 -*-
"""
@author: hubo
@project: bbb_utils
@file: singleton.py
@time: 2023/11/01
@desc: 通过@Singleton将类设置为单例模式
"""


class Singleton(object):
    """
    python版的单例模式
    使用例子
    @Singleton
    class Signle:
        def __init__(self):
            self.count = 0

        def add_count(self):
            self.count += 1
            return self.count


    print(Signle().add_count()) # 1
    print(Signle().add_count()) # 2
    """
    _INSTANCE = {}
    def __init__(self, cls):
        self.cls = cls

    def __call__(self, *args, **kwargs):
        instance = self._INSTANCE.get(self.cls, None)
        if not instance:
            instance = self.cls(*args, **kwargs)
            self._INSTANCE[self.cls] = instance
        return instance

    def __getattr__(self, key):
        return getattr(self.cls, key, None)





# -*- coding: utf-8 -*-
"""
@author: hubo
@project: fastframe
@file: oauth_server
@time: 2024/5/23 16:55
@desc:
"""
import uvicorn
from typing import Annotated

from fastoauth import OAuth2, Token, AuthMiddleware, get_current_user
from fastoauth import Token
from fastapi import Depends, FastAPI, HTTPException, status, Request
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from pydantic import BaseModel


# 在这里调整配置内容。详细说明参考R
# EADME
# OAuth2.SECRET_KEY =
# OAuth2.EXPIRE_SECONDS = 60*60*2
# OAuth2.REDIS_URL = 'redis://localhost:6379/oauth2'


class User(BaseModel):
    """
    创建一个User用来指示Token中包含的数据。
    生成文档用，可能为了灵活性，后续要做修改
    """
    username: str


oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/oauth2/token")
app = FastAPI(docs_url="/oauth2/docs")
app.add_middleware(AuthMiddleware)


@app.post("/oauth2/token", response_model=Token)
async def login_for_access_token(
    form_data: Annotated[OAuth2PasswordRequestForm, Depends()]
):
    """校验账号密码，获取access_token"""
    username = form_data.username
    password = form_data.password
    # 这里添加校验代码
    if username == 'johndoe' and password == 'secret':
        # 这里添加要写入到token中的数据
        return OAuth2.create_access_token({"username": username})
    raise HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Incorrect username or password",
        headers={"WWW-Authenticate": "Bearer"},
    )


@app.post("/oauth2/logout")
async def remove_token(token: Annotated[str, Depends(oauth2_scheme)]):
    """删除服务端缓存的token"""
    OAuth2.remove_token(token)
    return {}


@app.post("/oauth2/refresh_token", response_model=Token)
async def refresh_token(token: Annotated[str, Depends(oauth2_scheme)]):
    """在token过期之前刷新token，无须重新校验账号密码"""
    return OAuth2.refresh_token(token)


@app.get("/oauth2/me", response_model=User)
async def read_users_me(token: Annotated[str, Depends(oauth2_scheme)]):
    """获取缓存在token中用户信息"""
    # return OAuth2.get_token_user(token)
    return get_current_user()


def main():
    uvicorn.run(app, host="0.0.0.0", port=8000, reload=False)


if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
"""
@author: hubo
@project: bb_fast_oauth2
@file: schemas.py
@time: 2023/11/01
@desc: oauth2请求和返回的数据模型
"""
from pydantic import BaseModel


class Token(BaseModel):
    access_token: str
    refresh_token: str    # 直接使用access_token进行刷新
    token_type: str
    expires_in: int




# -*- coding: utf-8 -*-
"""
@author: hubo
@project: bb_fast_oauth2
@file: oauth.py
@time: 2023/11/01
@desc: 对oauth2中一些基本操作的封装
"""
import copy

import json
import time
from calendar import timegm
from contextvars import ContextVar
from datetime import datetime
from jose import JWTError, jwt
from fastapi import HTTPException, status, Request
import secrets
import uuid

from starlette.middleware.base import BaseHTTPMiddleware

from .schemas import Token
from ._redis import _Redis

current_token: ContextVar = ContextVar("current_token", default=None)


def get_current_user():
    token = current_token.get()
    if not token:
        return None
    return OAuth2.get_token_user(token)


class OAuth2:
    SECRET_KEY = secrets.token_bytes(32).hex()  # SECRET_KEY进行加密用的key
    EXPIRE_SECONDS = 60*60*2    # token有效期，单位秒。默认两个小时
    REDIS_URL = None            # 指定redis，格式redis://user:password@localhost:6379/db
    ALGORITHM = "HS256"         # 加密Token的算法

    @staticmethod
    def create_access_token(data: dict, expire_seconds=None):
        to_encode = data.copy()
        expire = (timegm(datetime.utcnow().utctimetuple()))
        if not expire_seconds:
            expire_seconds = OAuth2.EXPIRE_SECONDS
        expire += expire_seconds
        to_encode.update({"exp": expire})
        access_token, refresh_token = OAuth2._encode(to_encode)
        # to_encode.update({"_refresh": True})
        # refresh_token = OAuth2._encode(to_encode)
        # 暂不使用refresh_token，直接使用access_token进行刷新。一般refresh_token的有效期是access_token有效期的2倍
        return Token(access_token=access_token, refresh_token=refresh_token, token_type='bearer', expires_in=expire_seconds)

    @staticmethod
    def refresh_token(token, expire_minutes=None):
        payload = OAuth2._decode(token)
        if "_refresh" not in payload:
            # 只有refresh_token才可以刷新token
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Not a refresh token")
        # 删除refresh项
        del payload["_refresh"]
        return OAuth2.create_access_token(payload, expire_minutes)

    @staticmethod
    def remove_token(token):
        """如果启用redis，后端支持logout"""
        if _Redis().enabled_from(OAuth2.REDIS_URL):
            data = OAuth2._decode(token)
            # 配置redis，删除token
            _Redis().delete(token)
            _Redis().delete(data["refresh_token"])

    @staticmethod
    def get_token_user(token):
        """得到token中的数据"""
        data = OAuth2._decode(token)
        if "_refresh" in data:
            # refresh_token
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Not a access token")
        return data

    @staticmethod
    def _encode(data):
        token_data = copy.copy(data)
        """如果未启用redis，将整个数据加密作为token；如果启用了redis，把数据存在redis，并返回一个相对简短的加密的key作为token"""
        if _Redis().enabled_from(OAuth2.REDIS_URL):
            # 启用了redis，只加密一个唯一键作为检索的依据
            token = jwt.encode({'uid': str(uuid.uuid1())}, secrets.token_bytes(32).hex(), algorithm=OAuth2.ALGORITHM)
            refresh_token = jwt.encode({'uid': str(uuid.uuid1())}, secrets.token_bytes(32).hex(), algorithm=OAuth2.ALGORITHM)
            token_data["refresh_token"] = refresh_token
            _Redis().set(token, json.dumps(token_data), expire_in=int(time.time()) + OAuth2.EXPIRE_SECONDS)
            token_data.update({"_refresh": True})
            _Redis().set(refresh_token, json.dumps(token_data), expire_in=int(time.time()) + OAuth2.EXPIRE_SECONDS)
            return token, refresh_token
        else:
            # 未启用redis，将整个token_data加密，将加密后的字符串作为access_token
            token = jwt.encode(token_data, OAuth2.SECRET_KEY, algorithm=OAuth2.ALGORITHM)
            token_data.update({"_refresh": True})
            refresh_token = jwt.encode(token_data, OAuth2.SECRET_KEY, algorithm=OAuth2.ALGORITHM)
            return token, refresh_token

    @staticmethod
    def _decode(token):
        """如果未启用redis，token解密后就是完整数据；如果启用了redis，token只是redis的一个key，从redis取回完整数据"""
        if _Redis().enabled_from(OAuth2.REDIS_URL):
            # 启用了redis，token只是Redis里的检索依据
            data = _Redis().get(token)
            if not data:
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    detail="Could not validate credentials",
                    headers={"WWW-Authenticate": "Bearer"},
                )
            return json.loads(data)
        else:
            # 未启用redis，将整个token解密，得到原始数据
            try:
                return jwt.decode(token, OAuth2.SECRET_KEY, algorithms=[OAuth2.ALGORITHM])
            except JWTError:
                raise HTTPException(
                        status_code=status.HTTP_401_UNAUTHORIZED,
                        detail="Could not validate credentials",
                        headers={"WWW-Authenticate": "Bearer"},
                    )


class AuthMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        authorization_header = request.headers.get('Authorization')
        if authorization_header:
            # token 格式为 "Bearer <token>"
            token = authorization_header.split(" ")[1]
            current_token.set(token)
        response = await call_next(request)
        return response





# -*- coding: utf-8 -*-
"""
@author: hubo
@project: fastframe
@file: __init__.py
@time: 2024/5/21 16:19
@desc:
"""
from fastoauth.oauth import OAuth2, AuthMiddleware, get_current_user
from fastoauth.schemas import Token
